# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import logging
import struct

from gevent import monkey

monkey.patch_socket()

from contextlib import closing, contextmanager

from pymodbus.client.sync import ModbusTcpClient as SyncModbusClient
from pymodbus.constants import Defaults
from pymodbus.exceptions import ConnectionException, ModbusException, ModbusIOException
from pymodbus.pdu import ExceptionResponse
from volttron.driver.base.driver_locks import socket_lock
from volttron.driver.base.interfaces import (
    BaseInterface,
    BaseRegister,
    BasicRevert,
    DriverInterfaceError,
)
from volttron.utils.logs import setup_logging

setup_logging()
_log = logging.getLogger(__name__)


class ModbusInterfaceException(ModbusException):
    pass


class ModbusRegisterBase(BaseRegister):

    def __init__(self,
                 address,
                 register_type,
                 read_only,
                 pointName,
                 units,
                 description='',
                 slave_id=0):
        super(ModbusRegisterBase, self).__init__(register_type,
                                                 read_only,
                                                 pointName,
                                                 units,
                                                 description=description)
        self.address = address
        self.slave_id = slave_id


class ModbusBitRegister(ModbusRegisterBase):

    def __init__(self,
                 address,
                 type_string,
                 pointName,
                 units,
                 read_only,
                 mixed_endian=False,
                 description='',
                 slave_id=0):
        super(ModbusBitRegister, self).__init__(address,
                                                "bit",
                                                read_only,
                                                pointName,
                                                units,
                                                description=description,
                                                slave_id=slave_id)

        self.python_type = bool

    def parse_value(self, starting_address, bit_stream):
        # find the bytes we care about
        index = (self.address - starting_address)
        return bit_stream[index]

    def get_register_count(self):
        return 1

    def get_state(self, client):
        response_bits = client.read_discrete_inputs(self.address, unit=self.slave_id) if self.read_only else \
            client.read_coils(self.address, unit=self.slave_id)
        if response_bits is None:
            raise ModbusInterfaceException("pymodbus returned None")
        return response_bits.bits[0]

    def set_state(self, client, value):
        if not self.read_only:
            response = client.write_coil(self.address, value, unit=self.slave_id)
            if response is None:
                raise ModbusInterfaceException("pymodbus returned None")
            if isinstance(response, ExceptionResponse):
                raise ModbusInterfaceException(str(response))
            return response.value
        return None


class ModbusByteRegister(ModbusRegisterBase):

    def __init__(self,
                 address,
                 type_string,
                 pointName,
                 units,
                 read_only,
                 mixed_endian=False,
                 description='',
                 slave_id=0):
        super(ModbusByteRegister, self).__init__(address,
                                                 "byte",
                                                 read_only,
                                                 pointName,
                                                 units,
                                                 description=description,
                                                 slave_id=slave_id)

        try:
            self.parse_struct = struct.Struct(type_string)
        except struct.error:
            raise ValueError("Invalid Modbus Register '" + type_string + "' for point " +
                             pointName)

        struct_types = [
            type(x) for x in self.parse_struct.unpack(b'\x00' * self.parse_struct.size)
        ]

        if len(struct_types) != 1:
            raise ValueError("Invalid length Modbus Register '" + type_string + "' for point " +
                             pointName)

        self.python_type = struct_types[0]

        self.mixed_endian = mixed_endian
        self.modbus_register_size = 2
        self.pymodbus_register_struct = struct.Struct('>H')

    @contextmanager
    def modbus_client(self, address, port):
        with socket_lock():
            with closing(SyncModbusClient(address, port)) as client:
                yield client

    def get_register_count(self):
        return self.parse_struct.size // self.modbus_register_size

    def parse_value(self, starting_address, byte_stream):
        # find the bytes we care about
        index = (self.address - starting_address) * 2
        width = self.parse_struct.size

        target_bytes = byte_stream[index:index + width]
        if len(target_bytes) < width:
            raise ValueError('Not enough data to parse')

        if self.mixed_endian:
            register_values = []
            for i in range(0, len(target_bytes), self.pymodbus_register_struct.size):
                register_values.extend(self.pymodbus_register_struct.unpack_from(target_bytes, i))
            register_values.reverse()

            target_bytes = ""
            target_bytes = bytes.join(
                b'', [self.pymodbus_register_struct.pack(value) for value in register_values])
            # for value in register_values:
            #     target_bytes += PYMODBUS_REGISTER_STRUCT.pack(value).decode('utf-8')

        return self.parse_struct.unpack(target_bytes)[0]

    def get_state(self, client):
        if self.read_only:
            response = client.read_input_registers(self.address,
                                                   self.get_register_count(),
                                                   unit=self.slave_id)
        else:
            response = client.read_holding_registers(self.address,
                                                     self.get_register_count(),
                                                     unit=self.slave_id)

        if response is None:
            raise ModbusInterfaceException("pymodbus returned None")

        if self.mixed_endian:
            response.registers.reverse()

        response_bytes = response.encode()
        # skip the result count
        return self.parse_struct.unpack(response_bytes[1:])[0]

    def set_state(self, client, value):
        if not self.read_only:
            value_bytes = self.parse_struct.pack(value)
            register_values = []
            for i in range(0, len(value_bytes), self.pymodbus_register_struct.size):
                register_values.extend(self.pymodbus_register_struct.unpack_from(value_bytes, i))
            if self.mixed_endian:
                register_values.reverse()
            client.write_registers(self.address, register_values, unit=self.slave_id)
            return self.get_state(client)
        return None


class Modbus(BasicRevert, BaseInterface):

    def __init__(self, **kwargs):
        super(Modbus, self).__init__(**kwargs)
        self.register_ranges = {
            ('byte', True): [],
            ('byte', False): [],
            ('bit', True): [],
            ('bit', False): []
        }
        self.modbus_read_max = 100

    ##### Implemented abstract methods from BaseInterface

    def configure(self, config_dict, registry_config_str):
        self.slave_id = config_dict.get("slave_id", 0)
        self.ip_address = config_dict["device_address"]
        self.port = config_dict.get("port", Defaults.Port)
        self.parse_config(registry_config_str)

    def get_point(self, point_name):
        register = self.get_register_by_name(point_name)
        with self.modbus_client(self.ip_address, self.port) as client:
            try:
                result = register.get_state(client)
            except (ConnectionException, ModbusIOException, ModbusInterfaceException):
                result = None
        return result

    ##### Overriden methods from BaseInterface

    def insert_register(self, register):
        super(Modbus, self).insert_register(register)

        # MODBUS requires extra bookkeeping.
        register_type = register.get_register_type()
        register_range = self.register_ranges[register_type]
        register_count = register.get_register_count()

        # Store the range of registers for each point.
        start, end = register.address, register.address + register_count - 1
        register_range.append([start, end, [register]])

    ##### Uses implementation from BasicRevert for following BaseInterface abstract methods:
    ##### set_point, scrape_all, revert_all, revert_point

    ##### Implemented abstract methods from BasicRevert

    def _set_point(self, point_name, value):
        register = self.get_register_by_name(point_name)
        if register.read_only:
            raise IOError("Trying to write to a point configured read only: " + point_name)

        with self.modbus_client(self.ip_address, self.port) as client:
            try:
                result = register.set_state(client, value)
            except (ConnectionException, ModbusIOException, ModbusInterfaceException) as ex:
                raise IOError("Error encountered trying to write to point {}: {}".format(
                    point_name, ex))
        return result

    def _scrape_all(self):
        result_dict = {}
        with self.modbus_client(self.ip_address, self.port) as client:
            try:

                result_dict.update(self.scrape_byte_registers(client, True))
                result_dict.update(self.scrape_byte_registers(client, False))

                result_dict.update(self.scrape_bit_registers(client, True))
                result_dict.update(self.scrape_bit_registers(client, False))
            except (ConnectionException, ModbusIOException, ModbusInterfaceException) as e:
                raise DriverInterfaceError("Failed to scrape device at " + self.ip_address + ":" +
                                           str(self.port) + " ID: " + str(self.slave_id) + str(e))

        return result_dict

    ##### Helper methods

    def scrape_byte_registers(self, client, read_only):
        result_dict = {}
        register_ranges = self.register_ranges[('byte', read_only)]

        read_func = client.read_input_registers if read_only else client.read_holding_registers

        for register_range in register_ranges:
            start, end, registers = register_range
            result = b''

            for group in range(start, end + 1, self.modbus_read_max):
                count = min(end - group + 1, self.modbus_read_max)
                response = read_func(group, count, unit=self.slave_id)
                if response is None:
                    raise ModbusInterfaceException("pymodbus returned None")
                if isinstance(response, ModbusException):
                    raise response
                response_bytes = response.encode()
                # Trim off length byte.
                result += response_bytes[1:]

            for register in registers:
                point = register.point_name
                value = register.parse_value(start, result)
                result_dict[point] = value

        return result_dict

    def scrape_bit_registers(self, client, read_only):
        result_dict = {}
        register_ranges = self.register_ranges[('bit', read_only)]

        for register_range in register_ranges:
            start, end, registers = register_range
            if not registers:
                return result_dict

            result = []

            for group in range(start, end + 1, self.modbus_read_max):
                count = min(end - group + 1, self.modbus_read_max)
                response = client.read_discrete_inputs(group, count, unit=self.slave_id) if read_only else \
                    client.read_coils(group, count, unit=self.slave_id)
                if response is None:
                    raise ModbusInterfaceException("pymodbus returned None")
                if isinstance(response, ModbusException):
                    raise response
                result += response.bits

            for register in registers:
                point = register.point_name
                value = register.parse_value(start, result)
                result_dict[point] = value

        return result_dict

    def parse_config(self, configDict):
        if configDict is None:
            return

        for regDef in configDict:
            # Skip lines that have no address yet.
            if not regDef['Volttron Point Name']:
                continue

            io_type = regDef['Modbus Register']
            bit_register = io_type.lower() == 'bool'
            read_only = regDef['Writable'].lower() != 'true'
            point_path = regDef['Volttron Point Name']
            address = int(regDef['Point Address'])
            description = regDef.get('Notes', '')
            units = regDef['Units']

            default_value = regDef.get("Default Value", '').strip()

            mixed_endian = regDef.get('Mixed Endian', '').strip().lower() == 'true'

            klass = ModbusBitRegister if bit_register else ModbusByteRegister
            register = klass(address,
                             io_type,
                             point_path,
                             units,
                             read_only,
                             mixed_endian=mixed_endian,
                             description=description,
                             slave_id=self.slave_id)

            self.insert_register(register)

            if not read_only:
                if default_value:
                    if isinstance(register, ModbusBitRegister):
                        try:
                            value = bool(int(default_value))
                        except ValueError:
                            value = default_value.lower().startswith(
                                't') or default_value.lower() == 'on'
                        self.set_default(point_path, value)
                    else:
                        try:
                            value = register.python_type(default_value)
                            self.set_default(point_path, value)
                        except ValueError:
                            _log.warning(
                                "Unable to set default value for {}, bad default value in configuration. "
                                "Using default revert method.".format(point_path))

                else:
                    _log.info(
                        "No default value supplied for point {}. Using default revert method.".
                        format(point_path))

        # Merge adjacent ranges for efficiency.
        self.merge_register_ranges()

    def merge_register_ranges(self):
        """
        Merges any adjacent registers for more efficient scraping. May only be called after all registers have been
        inserted."""
        for key, register_ranges in self.register_ranges.items():
            if not register_ranges:
                continue
            register_ranges.sort()
            result = []
            current = register_ranges[0]
            for register_range in register_ranges[1:]:
                if register_range[0] > current[1] + 1:
                    result.append(current)
                    current = register_range
                    continue

                current[1] = register_range[1]
                current[2].extend(register_range[2])

            result.append(current)

            self.register_ranges[key] = result

    @contextmanager
    def modbus_client(self, address, port):
        with socket_lock():
            with closing(SyncModbusClient(address, port)) as client:
                yield client
